import json

import pkg_resources
from marshmallow import Schema
from pyramid.path import DottedNameResolver

try:
    import yaml
    from apispec import APISpec, utils, yaml_utils
    from apispec.ext.marshmallow.common import (
        resolve_schema_cls,
        resolve_schema_instance,
    )
except ImportError:
    raise ImportError(
        "You must have the `apispec` package installed to use this feature.  "
        "You can install it with `pip install pyramid_marshmallow[openapi]."
    )


def schema_name_resolver(schema):
    cls = resolve_schema_cls(schema)
    instance = resolve_schema_instance(schema)
    name = cls.__name__
    if not cls.opts.register:
        # Unregistered schemas are put inline.
        return False
    if instance.only or instance.exclude:
        # If schema includes only select fields, treat it as nonce
        return False
    if name.endswith("Schema"):
        name = name[:-6] or name
    if instance.partial:
        name = "Partial" + name
    return name


def list_paths(introspector):
    for item in introspector.get_category("views"):
        spect = item["introspectable"]
        path = make_path(introspector, spect)
        if path is None:
            continue
        operations = dict()
        methods = spect["request_methods"] or ["GET"]
        if isinstance(methods, str):
            methods = [methods]
        for method in methods:
            method = method.lower()
            operations[method] = spect
        yield path, operations


def make_path(introspector, introspectable):
    if introspectable["route_name"]:
        route = introspector.get("routes", introspectable["route_name"])
        return route["pattern"]
    elif introspectable["context"]:
        context = introspectable["context"]
        base = getattr(context, "__path__", None)
        if base is None:
            return None
        else:
            return base + "/" + (introspectable["name"] or "")
    else:
        return None


def _schema(schema):
    if isinstance(schema, dict):
        return Schema.from_dict(schema)
    else:
        return schema


def split_docstring(docstring):
    """
    Split a docstring in half, delineated with a "---".  The first half is
    returned verbatim, the second half is parsed as YAML.

    """
    split_lines = utils.trim_docstring(docstring).split("\n")

    # Cut YAML from rest of docstring
    for index, line in enumerate(split_lines):
        line = line.strip()
        if line.startswith("---"):
            cut_from = index
            break
    else:
        cut_from = len(split_lines)

    summary = split_lines[0].strip() or None
    docs = "\n".join(split_lines[1:cut_from]).strip() or None
    yaml_string = "\n".join(split_lines[cut_from:])
    if yaml_string:
        parsed = yaml.safe_load(yaml_string)
    else:
        parsed = dict()
    return summary, docs, parsed


def set_request_body(spec, op, view):
    op["requestBody"] = {
        "content": {
            "application/json": {
                "schema": _schema(view["validate"]),
            },
        },
    }


def set_query_params(spec, op, view):
    op["parameters"].append(
        {
            "in": "query",
            "schema": _schema(view["validate"]),
        }
    )


def set_response_body(spec, op, view):
    op["responses"]["200"] = {
        "description": "",
        "content": {
            "application/json": {
                "schema": _schema(view["marshal"]),
            },
        },
    }


def set_url_params(spec, op, view):
    context = view["context"]
    if not context:
        return
    params = getattr(context, "__params__", [])
    for param in params:
        param["in"] = "path"
        param["required"] = True
    op["parameters"].extend(params)


def set_tag(spec, op, view):
    context = view["context"]
    if not context:
        return
    tag = getattr(context, "__tag__", None)
    if not tag:
        return
    if isinstance(tag, dict):
        # Cheating and using the private variable spec._tags
        if not any(x["name"] == tag["name"] for x in spec._tags):
            spec.tag(tag)
        tag_name = tag["name"]
    else:
        tag_name = tag
    op.setdefault("tags", []).append(tag_name)


def create_spec(registry, zone=None, merge=None):
    title = registry.settings.get("openapi.title", "Untitled")
    version = registry.settings.get("openapi.version", "0.0.0")
    name_resolver = DottedNameResolver()
    MarshmallowPlugin = name_resolver.maybe_resolve(
        registry.settings.get(
            "openapi.plugin", "apispec.ext.marshmallow.MarshmallowPlugin"
        )
    )
    marshmallow_plugin = MarshmallowPlugin(
        schema_name_resolver=schema_name_resolver
    )
    spec = APISpec(
        title=title,
        version=version,
        openapi_version="3.0.2",
        plugins=[marshmallow_plugin],
    )
    plugin_hook = registry.settings.get("openapi.plugin_hook")
    if plugin_hook:
        plugin_hook = name_resolver.maybe_resolve(plugin_hook)
        plugin_hook(registry, spec, marshmallow_plugin)
    for path, operations in list_paths(registry.introspector):
        final_ops = dict()
        for method, view in operations.items():
            if zone is not None and zone != view.get("api_zone"):
                continue
            summary, descr, user_op = split_docstring(view["callable"].__doc__)
            op = {
                "responses": dict(),
                "parameters": [],
            }
            if summary:
                op["summary"] = summary
            if descr:
                op["description"] = descr
            set_url_params(spec, op, view)
            if "validate" in view:
                if method == "get":
                    set_query_params(spec, op, view)
                else:
                    set_request_body(spec, op, view)
            if "marshal" in view:
                set_response_body(spec, op, view)
            set_tag(spec, op, view)
            final_op = utils.deepupdate(op, user_op)
            final_op = utils.deepupdate(final_op, view.get("api_spec", dict()))

            # We are required to have some response, so make one up.
            if not final_op["responses"]:
                final_op["responses"]["200"] = {
                    "description": "",
                }
            final_ops[method] = final_op
        spec.path(path, operations=final_ops)

    json = spec.to_dict()
    return _perform_merges(json, merge, registry.settings.get("openapi.merge"))


def _perform_merges(json, mergefile, merge_setting):
    # Perform merges
    if mergefile is None:
        merges = []
    elif isinstance(mergefile, str):
        merges = [mergefile]
    else:
        merges = mergefile
    if not merge_setting:
        pass
    elif isinstance(merge_setting, str):
        merges.extend(x.strip() for x in merge_setting.split(","))
    else:
        merges.extend(merge_setting)
    for mergefile in merges:
        json = merge(json, mergefile)
    return json


def merge(spec, mergefile):
    if ":" in mergefile:
        module, _, path = mergefile.partition(":")
        fh = pkg_resources.resource_stream(module, path)
    else:
        fh = open(mergefile)
    with fh:
        to_merge = yaml.safe_load(fh)
    return utils.deepupdate(spec, to_merge)


def generate_html(spec):
    data = json.dumps(spec, sort_keys=True)
    return HTML_TEMPLATE.format(
        title=spec["info"]["title"],
        version=spec["info"]["version"],
        spec=data,
    )


HTML_TEMPLATE = """
<!DOCTYPE html>
<html>
    <head>
        <title>{title} {version}</title>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400\
,700|Roboto:300,400,700" rel="stylesheet">
        <style>
            body {{
                margin: 0;
                padding: 0;
            }}
        </style>
    </head>
    <body>
        <div id="redoc"></div>
        <script type="text/json" id="spec">{spec}</script>
        <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.sta\
ndalone.js"></script>
        <script>
            window.addEventListener('load', function() {{
                var el = document.getElementById('redoc');
                var spec = JSON.parse(document.getElementById('spec').text);
                Redoc.init(spec, {{}}, el);
            }});
        </script>
    </body>
</html>
"""


def generate_yaml(spec):
    return yaml_utils.dict_to_yaml(spec)



from datetime import date as Date
from types import SimpleNamespace
from unittest.mock import Mock

import pytest
from marshmallow import Schema, ValidationError, fields
from pyramid.testing import DummyRequest
from webob.multidict import MultiDict

from pyramid_marshmallow import view_validator


class AlbumSchema(Schema):
    title = fields.Str()
    release_date = fields.Date()


@pytest.fixture
def view():
    return Mock()


@pytest.fixture
def wrapped(view):
    info = SimpleNamespace(
        options={
            "validate": AlbumSchema(),
        }
    )
    return view_validator(view, info)


def test_no_validate():
    view = Mock()
    info = SimpleNamespace(options={})
    assert view_validator(view, info) is view


def test_validate(wrapped, view):
    request = DummyRequest()
    request.method = "POST"
    request.json_body = {
        "title": "Hunky Dory",
        "release_date": "1971-12-17",
    }
    context = object()
    wrapped(context, request)
    view.assert_called_once_with(context, request)
    assert request.data == {
        "title": "Hunky Dory",
        "release_date": Date(1971, 12, 17),
    }


def test_validate_get(wrapped, view):
    request = DummyRequest()
    request.method = "GET"
    request.GET = MultiDict(
        {
            "title": "Hunky Dory",
            "release_date": "1971-12-17",
        }
    )
    context = object()
    wrapped(context, request)
    view.assert_called_once_with(context, request)
    assert request.data == {
        "title": "Hunky Dory",
        "release_date": Date(1971, 12, 17),
    }


def test_validate_error(wrapped, view):
    """
    Technically not necessary, since we just let marshmallow use its own error
    handling.  However, our documentation makes claims on how this works, so we
    test it.

    """
    request = DummyRequest()
    request.method = "POST"
    request.json_body = {
        "title": "Hunky Dory",
        "release_date": "1971-14-17",
    }
    context = object()
    with pytest.raises(ValidationError) as exc:
        wrapped(context, request)
    assert exc.value.messages == {"release_date": ["Not a valid date."]}
    assert exc.value.normalized_messages() == {
        "release_date": ["Not a valid date."]
    }



import argparse
import json
from importlib import import_module

from pyramid.paster import get_app

from .spec import create_spec, generate_html, generate_yaml

parser = argparse.ArgumentParser()
parser.add_argument(
    "app",
    nargs="?",
    help="The app module.  In the format `module:variable`.",
)
parser.add_argument(
    "--ini",
    help="If specified, load the app via Paste from an ini file.",
)
parser.add_argument(
    "--zone",
    help=(
        "The API zone to generate spec for.  See documentation for more "
        "details."
    ),
)
parser.add_argument(
    "--merge",
    help="A YAML file to merge with the generated spec.",
    nargs="*",
)
parser.add_argument(
    "--format",
    help='The output, one of "json", "yaml", or "html".',
    default="json",
)
parser.add_argument(
    "--output",
    help="The file to output to.",
    default="-",
)


def generate():
    args = parser.parse_args()
    if args.app and args.ini:
        raise ValueError("Cannot specify both [app] and --ini.")
    elif args.app:
        app = import_app(args.app)
    elif args.ini:
        app = get_app(args.ini)
    else:
        raise ValueError("Must specify one of [app] or --ini.")
    spec_json = create_spec(app.registry, zone=args.zone, merge=args.merge)
    if args.format == "json":
        output = json.dumps(spec_json, sort_keys=True)
    elif args.format == "yaml":
        output = generate_yaml(spec_json)
    elif args.format == "html":
        output = generate_html(spec_json)
    else:
        raise ValueError('Format must be one of "json", "yaml", or "html".')
    if args.output == "-":
        print(output)  # noqa: T201
    else:
        with open(args.output, "wb") as fh:
            if isinstance(output, str):
                output = output.encode("utf8")
            fh.write(output)


def import_app(name):
    module_name, _, var_name = name.partition(":")
    module = import_module(module_name)
    return getattr(module, var_name)


if __name__ == "__main__":
    generate()



from marshmallow import Schema, ValidationError, fields
from pyramid.response import Response
from pyramid.viewderivers import VIEW

__all__ = [
    "ValidationError",
]


def includeme(config):
    config.add_view_deriver(view_validator)
    config.add_view_deriver(view_marshaller, under="rendered_view", over=VIEW)
    config.add_view_deriver(view_api_spec)


def make_schema(schema=None, **kwargs):
    """
    Create a schema from a dictionary.

    **Deprecated**:  Use `Schema.from_dict` instead.

    """
    return Schema.from_dict(schema or kwargs)


def process_schema(schema):
    """
    Handle a schema passed in as a view deriver, creating a nonce schema if a
    dictionary.

    """
    if schema is None:
        return None
    elif isinstance(schema, Schema):
        return schema
    elif isinstance(schema, dict):
        return Schema.from_dict(schema)()
    else:
        raise TypeError("Schema is of invalid type.")


def view_validator(view, info):
    schema = process_schema(info.options.get("validate"))
    if schema is None:
        return view

    def wrapped(context, request):
        if request.method == "GET":
            data = dict()
            for k, v in request.GET.items():
                field = schema.fields.get(k)
                if isinstance(field, fields.List):
                    data.setdefault(k, []).append(v)
                else:
                    data[k] = v
        else:
            data = request.json_body
        request.data = schema.load(data)
        return view(context, request)

    return wrapped


view_validator.options = ("validate",)


def view_marshaller(view, info):
    schema = process_schema(info.options.get("marshal"))
    if schema is None:
        return view

    def wrapped(context, request):
        output = view(context, request)
        if isinstance(output, Response):
            return output
        else:
            return schema.dump(output)

    return wrapped


view_marshaller.options = ("marshal",)


def view_api_spec(view, info):
    return view


view_api_spec.options = ("api_spec", "api_zone")



from datetime import date as Date
from types import SimpleNamespace
from unittest.mock import Mock

import pytest
from marshmallow import Schema, fields
from pyramid.httpexceptions import HTTPNoContent

from pyramid_marshmallow import view_marshaller


class AlbumSchema(Schema):
    title = fields.Str()
    release_date = fields.Date()


@pytest.fixture
def wrap_view():
    def wrap(view):
        info = SimpleNamespace(
            options={
                "marshal": AlbumSchema(),
            }
        )
        return view_marshaller(view, info)

    return wrap


def test_marshal(wrap_view):
    unwrapped = Mock(
        return_value={
            "title": "Hunky Dory",
            "release_date": Date(1971, 12, 17),
        }
    )
    view = wrap_view(unwrapped)
    assert view("context", "request") == {
        "title": "Hunky Dory",
        "release_date": "1971-12-17",
    }
    unwrapped.assert_called_once_with("context", "request")


def test_marshal_http_response(wrap_view):
    unwrapped = Mock(return_value=HTTPNoContent())
    view = wrap_view(unwrapped)
    resp = view("context", "request")
    assert isinstance(resp, HTTPNoContent)



def test_hello_world(app):
    "Just to make sure our sample application is working properly."
    r = app.get("/hello")
    assert r.text == "Hello World!"


def test_traversal_hello(app):
    r = app.get("/album/hello")
    assert r.text == "Hello World!"


def test_post_validate_integration(app):
    assert (
        app.post_json(
            "/validate",
            {
                "title": "Hunky Dory",
                "release_date": "1971-12-17",
            },
        ).text
        == "Hunky Dory"
    )


def test_get_validate_integration(app):
    assert (
        app.get(
            "/validate",
            {
                "title": "Hunky Dory",
                "release_date": "1971-12-17",
            },
        ).text
        == "Hunky Dory"
    )


def test_get_validate_list_integration(app):
    assert (
        app.get(
            "/validate-list",
            {
                "title": "Hunky Dory",
                "artists": ["Bowie", "Wowie"],
            },
        ).text
        == "Bowie, Wowie"
    )


def test_marshal_integration(app):
    assert app.get("/marshal").json == {
        "title": "Hunky Dory",
        "release_date": "1971-12-17",
    }



import functools
import json

from pyramid.response import Response
from zope.interface import Interface, implementer

from .spec import create_spec, generate_html, generate_yaml


def includeme(config):
    config.registry.registerUtility(
        SpecGenerator(config.registry),
        ISpecGenerator,
    )
    config.add_directive("add_openapi_json_view", add_openapi_json_view)
    config.add_directive("add_openapi_html_view", add_openapi_html_view)
    config.add_directive("add_openapi_yaml_view", add_openapi_yaml_view)


class ISpecGenerator(Interface):
    """
    Interface for a simple callable that returns the spec.

    """

    def __call__(zone, merge): ...


@implementer(ISpecGenerator)
class SpecGenerator:
    """
    Generate and cache specs for the given registry.

    """

    def __init__(self, registry):
        self.registry = registry

    def __call__(self, zone, merge):
        return self._call(zone, tuple(merge) if merge else tuple())

    @functools.lru_cache
    def _call(self, zone, merge):
        return create_spec(self.registry, zone=zone, merge=merge)


def _inject_params(view, zone=None, merge=None):
    """
    Add `zone` and `merge` arguments to given view callable.

    Originally I was using `functools.partial` directly, but Pyramid would
    complain unless I added `functools.update_wrapper`.

    """
    wrapper = functools.partial(view, zone=zone, merge=merge)
    functools.update_wrapper(wrapper, view)
    return wrapper


def add_openapi_json_view(config, zone=None, merge=None, *args, **kwargs):
    """
    Add a view that returns the JSON spec with the given zone and mergefiles.
    Additional args are passed to `Configuration.add_view`.

    """
    config.add_view(
        _inject_params(json_view, zone=zone, merge=merge),
        *args,
        **kwargs,
    )


def json_view(request, zone=None, merge=None):
    generator = request.registry.getUtility(ISpecGenerator)
    indent = 2 if "pretty" in request.GET else None
    body = json.dumps(generator(zone, merge), indent=indent)
    return Response(
        body=body,
        content_type="application/json",
        charset="utf-8",
    )


def add_openapi_html_view(config, zone=None, merge=None, *args, **kwargs):
    """
    Add a view that returns a redoc page for a spec with the given zone and
    mergefiles.  Additional args are passed to `Configuration.add_view`.

    """
    config.add_view(
        _inject_params(html_view, zone=zone, merge=merge),
        *args,
        **kwargs,
    )


def html_view(request, zone=None, merge=None):
    generator = request.registry.getUtility(ISpecGenerator)
    body = generate_html(generator(zone, merge))
    return Response(
        body=body,
        content_type="text/html",
    )


def add_openapi_yaml_view(config, zone=None, merge=None, *args, **kwargs):
    """
    Add a view that returns a YAML spec with the given zone and mergefiles.
    Additional args are passed to `Configuration.add_view`.

    """
    config.add_view(
        _inject_params(yaml_view, zone=zone, merge=merge),
        *args,
        **kwargs,
    )


def yaml_view(request, zone=None, merge=None):
    generator = request.registry.getUtility(ISpecGenerator)
    body = generate_yaml(generator(zone, merge))
    return Response(
        body=body,
        content_type="text/yaml",
    )



from datetime import date as Date

from marshmallow import Schema, fields

from pyramid_marshmallow import make_schema, process_schema


class AlbumSchema(Schema):
    title = fields.Str()
    release_date = fields.Date()


def test_process_schema_null():
    assert process_schema(None) is None


def test_process_schema_schema():
    schema = AlbumSchema()
    assert process_schema(schema) is schema


def test_process_schema_dict():
    schema = process_schema(
        {
            "title": fields.Str(),
            "release_date": fields.Date(),
        }
    )
    assert isinstance(schema, Schema)
    assert schema.load(
        {
            "title": "Hunky Dory",
            "release_date": "1971-12-17",
        }
    ) == {
        "title": "Hunky Dory",
        "release_date": Date(1971, 12, 17),
    }


def test_make_schema_dict():
    _Schema = make_schema(
        {
            "title": fields.Str(),
            "release_date": fields.Date(),
        }
    )
    assert issubclass(_Schema, Schema)
    assert _Schema().load(
        {
            "title": "Hunky Dory",
            "release_date": "1971-12-17",
        }
    ) == {
        "title": "Hunky Dory",
        "release_date": Date(1971, 12, 17),
    }


def test_make_schema_kwargs():
    _Schema = make_schema(
        title=fields.Str(),
        release_date=fields.Date(),
    )
    assert issubclass(_Schema, Schema)
    assert _Schema().load(
        {
            "title": "Hunky Dory",
            "release_date": "1971-12-17",
        }
    ) == {
        "title": "Hunky Dory",
        "release_date": Date(1971, 12, 17),
    }



import argparse
from importlib import import_module
from wsgiref.simple_server import make_server

from pyramid.config import Configurator
from pyramid.paster import get_app

from . import ISpecGenerator, SpecGenerator

parser = argparse.ArgumentParser()
parser.add_argument(
    "app",
    nargs="?",
    help="The app module.  In the format `module:variable`.",
)
parser.add_argument(
    "--ini",
    help="If specified, load the app via Paste from an ini file.",
)
parser.add_argument(
    "--host",
    help="The host to bind to.  Defaults to 0.0.0.0.",
    default="0.0.0.0",
)
parser.add_argument(
    "--port",
    type=int,
    help="The port to bind to.  Defaults to 8080.",
    default=8080,
)
parser.add_argument(
    "--zone",
    help=(
        "The API zone to generate spec for.  See documentation for more "
        "details.",
    ),
)
parser.add_argument(
    "--merge",
    help="A YAML file to merge with the generated spec.",
    nargs="*",
)


def serve():
    args = parser.parse_args()
    if args.app and args.ini:
        raise ValueError("Cannot specify both [app] and --ini.")
    elif args.app:
        app = import_app(args.app)
    elif args.ini:
        app = get_app(args.ini)
    else:
        raise ValueError("Must specify one of [app] or --ini.")
    wsgi_app = create_wsgi_app(args, app.registry)
    server = make_server(args.host, args.port, wsgi_app)
    print(f"Starting server on { args.host }:{ args.port }")  # noqa: T201
    server.serve_forever()


def create_wsgi_app(args, registry):
    config = Configurator()
    config.include("pyramid_marshmallow.openapi")
    config.registry.registerUtility(SpecGenerator(registry), ISpecGenerator)
    config.add_route("index", "/")
    config.add_openapi_html_view(
        zone=args.zone,
        merge=args.merge,
        route_name="index",
    )
    config.add_route("json", "/spec.json")
    config.add_openapi_json_view(
        zone=args.zone,
        merge=args.merge,
        route_name="json",
    )
    config.add_route("yaml", "/spec.yaml")
    config.add_openapi_yaml_view(
        zone=args.zone,
        merge=args.merge,
        route_name="yaml",
    )
    return config.make_wsgi_app()


def import_app(name):
    module_name, _, var_name = name.partition(":")
    module = import_module(module_name)
    return getattr(module, var_name)


if __name__ == "__main__":
    serve()
